/* colinterface.c 
     Interface to the collaborative filtering system
     by Dave Maltz 10/93

  history:
  10/25/93 created

*/

#include "colinterface.h"
#include "collabdata.h"
#include "collabtcp.h"
#include "colinternal.h"

#include <stdio.h>
#include <string.h>
#include <pwd.h>   	/* porting problems??? */
#include <time.h>   /* porting problems??? */


/**** Internal Structure defs *******/

struct _group_votes {
  char			*group_name;
  struct _group_votes	*next_group;
  vote_t		*votes;
  vote_t		*current_vote;
};
typedef struct _group_votes group_votes;

/* different types of internal_magic */
#define OUT_VOTE_DATA 1
struct out_vote_data {
  char			type;
  char			anon;		/* = 1 if vote is to be anon */
};

/******* Internal Globals ***********/

static group_votes *outgoing_votes = NULL;
/* pointer to list of new votes cast by user in this session */

static group_votes *current_outgoing_group = NULL;
/* pointer to the last group in which a vote was cast */

/****** Internal Procedures *********/

static group_votes*  find_group_record(/* name,group_list */);
static group_votes* create_new_group_record(/* group_name */);
static void destroy_groups( /* group_votes *g */);

static void send_votes_by_tcp(/* groups */);
static void send_votes_by_mail(/* groups */);

/******* Internal Grottiness *********/
#define allocate(type) ((type *) collab_getmemory(sizeof(type)))

/*========================================================================
  Handle votes
  return status: 0 = failed, 1 = successful
------------------------------------------------------------------------*/
static vote_t *
  create_or_modify_vote(message_id, value, group_name)
char *message_id;
int value;
char *group_name;
/* Actually find out if we have a vote for the article already, if so, 
   mutate it to be like the user wants now, otherwise create the vote */
{
  vote_t *previous_vote;
  vote_data_t *temp;
  group_votes *group;
  
  if (current_outgoing_group == NULL || 
      strcasecmp(current_outgoing_group->group_name,group_name) != 0)
    {
      group = find_group_record(group_name, outgoing_votes);
      if (group == NULL)
	{
	  group = create_new_group_record(group_name);
	  group->next_group = outgoing_votes;
	  outgoing_votes = current_outgoing_group = group;
	}
      else current_outgoing_group = group;
    }
  
  if (current_outgoing_group->current_vote == NULL ||
      strcmp(current_outgoing_group->current_vote->message_id , message_id) != 0)
    {
      previous_vote = collab_find_previous_vote_record(message_id, 
						current_outgoing_group->current_vote);
      if (previous_vote == NULL)
	{
	  previous_vote = collab_create_new_vote_record(message_id);
	  previous_vote->internal_magic = allocate(struct out_vote_data);
	  ((struct out_vote_data *) previous_vote->internal_magic)->type 
	    = OUT_VOTE_DATA;
	  previous_vote->next_vote = current_outgoing_group->votes;
	  current_outgoing_group->votes = previous_vote;
	  current_outgoing_group->current_vote = previous_vote;
	}
      else
	current_outgoing_group->current_vote = previous_vote;
    }

  temp = current_outgoing_group->current_vote->vote_data;
  temp->num_high = 0;
  temp->num_med = 0;
  temp->num_low = 0;
  temp->num_terrible = 0;
  temp->stuff = NULL;

  switch(value) {
  case HIGH_VOTE:
    temp->num_high = 1;
    break;
  case MED_VOTE:
    temp->num_med = 1;
    break;
  case LOW_VOTE:
    temp->num_low = 1;
    break;
  case HATEIT_VOTE:
    temp->num_terrible = 1;
    break;
  default:
    break;
  }

  return current_outgoing_group->current_vote;
}

int 
  collab_enter_vote_anonymously(message_id, vote, group_name)
char *message_id;
int vote;
char *group_name;
{
  vote_t *temp;

  if (message_id == NULL) 
    return 0; /* this should never happen (something wierd is up...) */
  
  temp = create_or_modify_vote(message_id,vote,group_name);
  ((struct out_vote_data *) temp->internal_magic)->anon = 1;
  return 1;
}

char collab_user_name[NAME_SIZE];    /* K&R claim that this will be init'd to 0 */

int 
  collab_enter_vote(message_id, vote, group_name)
char *message_id;
int vote;
char *group_name;
{
  vote_t *temp;

  if (message_id == NULL) 
    return 0; /* this should never happen (something wierd is up...) */
  
  temp = create_or_modify_vote(message_id,vote,group_name);
  ((struct out_vote_data *) temp->internal_magic)->anon = 0;

  if (collab_user_name[0] == (char) 0)
    {
      struct passwd *pwd;
      pwd = getpwuid(getuid());
      (void) strncpy(collab_user_name, pwd->pw_gecos, NAME_SIZE);
      collab_user_name[NAME_SIZE -1] = '\0';
    }

  temp->vote_data->stuff = collab_user_name;
  return 1;
}

int 
  collab_send_off_votes(vote_file)
char *vote_file; /* pathname of vote file */
/* take the outgoing_votes and deliver them to the vote server */
/* add in stuff to allow writing to a file as well */
{
  if (outgoing_votes == NULL) return 0; /* no votes were cast... */
  
  if (vote_file != NULL) issue_error("sorry, we can't handle vote files yet");

#ifdef MAIL_VOTES
  send_votes_by_mail(outgoing_votes);
#endif /*  MAIL_VOTES */

#ifdef TCP_COMM
  send_votes_by_tcp(outgoing_votes);
#endif /* TCP_COMM */

  destroy_groups(outgoing_votes);
  return 1;
}

/*========================================================================
  Helper procs to send off vote data
------------------------------------------------------------------------*/
#define BUFLEN 200     /* the buflen is the maximum length that 
			  the message_id + votes + extra strings can be */
#ifdef MAIL_VOTES
static void
  send_votes_by_mail(groups)
group_votes *groups;
{
  FILE *vote_output;
  int status;
  group_votes *group_scan;
  vote_t *vote_scan;
  int size;
  char buf[BUFLEN];
  time_t tloc = time((time_t *) NULL);
  char *date = ctime(&tloc);

  vote_output = popen(VOTE_MAILER,"w");
  if (vote_output == NULL)
    {
      issue_error("Couldn't mail votes");
      return;
    }

  fprintf(vote_file,"to: VOTE_SERVER_ADDR \n");
  fprintf(vote_file,"cc: VOTE_CC_LIST \n"); 
  fprintf(vote_file,"subject: vote \n");
  fprintf(vote_file,"------------ \n");
  fprintf(vote_file,"+\n");
  fprintf(vote_file,"-- nn -- %s -- %s \n", date);
  free(date);

  for(group_scan = outgoing_votes; group_scan != NULL; 
      group_scan = group_scan->next_group)
    {
      fprintf(vote_output,"GROUP: %s \n", group_scan->group_name);
      
      for(vote_scan = group_scan->votes ; vote_scan != NULL;
	  vote_scan = vote_scan->next_vote)
	{
	  status = collab_parse_vote_into_string(buf, BUFLEN, vote_scan, &size);
	  if (status == 0) continue;   /* the vote didn't parse */
	  fprintf(vote_output,"%s\n",buf);
	} /* end for vote_scan */
    } /* end for group_scan */

  fprintf(vote_file,"+\n");
  pclose(vote_output);
}
#endif /* MAIL_VOTES */

#ifdef TCP_COMM
static void
  send_votes_by_tcp(groups)
group_votes *groups;
{
  int status;
  group_votes *group_scan;
  vote_t *vote_scan;
  int size;
  char buf[BUFLEN];

  /* send the server our announcement */
  (void) strcpy(buf,"VOTES\n");
  collabtcp_send_buf(buf,strlen(buf));

  /* send the votes */
  for(group_scan = groups; group_scan != NULL; 
      group_scan = group_scan->next_group)
    {
      sprintf(buf,"GROUP: %s \n", group_scan->group_name);
      status = collabtcp_send_buf(buf,strlen(buf));
      if(status == 0) continue; /* we've lost this group */

      for(vote_scan = group_scan->votes ; vote_scan != NULL;
	  vote_scan = vote_scan->next_vote)
	{
	  status = collab_parse_vote_into_string(buf, BUFLEN, vote_scan, &size);
	  if (status == 0) continue;   /* the vote didn't parse */
	  buf[size] = '\n';  /* overwrite the \0 */
	  collabtcp_send_buf(buf,size+1); /* we give size+1 to send \n */  

	} /* end for vote_scan */
    } /* end for group_scan */

  /* send the closing */
  (void) strcpy(buf, ".\n");
  collabtcp_send_buf(buf,strlen(buf));    
}
#endif /* TCP_COMM */


/*========================================================================
  Data structure for controlling multiple sources
------------------------------------------------------------------------*/
struct source {
  char 		source_type;
  char		*remote_name;	  /* URL for the source */
  char		*local_file_name; /* name of local cache of remote source */
};

#define Source_Server 1
#define Source_File   2

/*========================================================================
  Obtaining information from a source (hidden functions)
------------------------------------------------------------------------*/
static vote_t *
  get_votes_from_source(group_name, src)
char *group_name;
struct source *src;
/* Obtain votes posted in the given source for the given group.
   Allocate new memory for these votes.  Returns NULL if the operation
   fails at some point
*/
{
  vote_t *votes;

  if (src->source_type == Source_Server)
    {
      /* check to see if we have them cached somewhere */
      /*   to be implemented */
      
      /* if not, then get them from the vote server */
      votes = collabtcp_get_votes_via_tcp(group_name);
      
      /* stash the votes into the cache */
      /*   to be implemented */

      return votes;
    }
  else
    {
      issue_error("unknown source type -- notify maltz@parc.xerox.com");
    }
  return NULL;
}

char *
  get_available_groups_from_source(src, num_groups)
struct source *src;
int *num_groups;
/* Obtain a list of all the groups available from a source as a string of group
   names, each name seperated by a \n.
   Returns NULL if the operation fails at some point
*/
{
  if (src->source_type == Source_Server)
    {
      /* if not, then get them from the vote server */
      return collabtcp_get_group_list(num_groups);
    }
  else
    {
      issue_error("unknown source type -- notify maltz@parc.xerox.com");
    }
  return NULL;
}


/*========================================================================
  Obtaining Information from all the sources (exposed functions)
------------------------------------------------------------------------*/
vote_t *
  collab_get_votes_for_group(group_name)
char *group_name;
/* accumulate all the votes for this group from the sources currently in the
   source list, combine the votes and return them as linked list of vote_t
   returns NULL if something goes wrong
*/
{
  /* the current implementation of this ignores the sources and just goes
     to talk to the vote servers */
  
  struct source vote_server;
  
  vote_server.source_type = Source_Server;
  vote_server.remote_name = NULL;
  vote_server.local_file_name = NULL;

  return get_votes_from_source(group_name, &vote_server);
}

char* 
  collab_available_groups(num_groups)
int *num_groups;
/* return a new string of the group names sepperated by \n characters;
   the number of groups will be left in the loc pointed to by num_groups */
{
  /* the current implementation of this ignores the sources and just goes
     to talk to the vote server */
  
  struct source vote_server;
  
  vote_server.source_type = Source_Server;
  vote_server.remote_name = NULL;
  vote_server.local_file_name = NULL;

  return get_available_groups_from_source(&vote_server, num_groups);

}

/*========================================================================
  List utilities for the vote abstraction
------------------------------------------------------------------------*/
vote_t*
  collab_find_previous_vote_record(message_id,vote_list)
vote_t *vote_list;
char* message_id;
{   
  vote_t *scan;
  
  scan = vote_list;
  while (scan != NULL)
    {
      if (strcmp(scan->message_id,message_id) == 0) /* be case sensitive */
	return scan;
      scan = scan->next_vote;
    }
  return NULL;
}

vote_t* 
  collab_create_new_vote_record(message_id)
char *message_id;
/* allocate and init a new vote record */
{
  vote_t *temp;
  temp = allocate(vote_t);
  temp->message_id = (char *) collab_getmemory(strlen(message_id) + 1);
  (void) strcpy(temp->message_id, message_id);
  temp->vote_data = allocate(vote_data_t);
  temp->next_vote = NULL;
  temp->internal_magic = NULL;
  return temp;
}

void
  collab_destroy_votes(v)
vote_t *v;
{
  vote_t *temp;

  while(v != NULL)
    {
      free(v->message_id);
      if (v->vote_data->stuff != NULL) free(v->vote_data->stuff);
      free(v->vote_data);
      
      if (v->internal_magic != NULL)
	{
	  /* we may one day have to check to see what type of internal magic
	     is about to be freed */
	  free(v->internal_magic);
	}
      temp = v;
      v = v->next_vote;
      free(temp);
    }
}

/*========================================================================
  List utilities for the group_votes abstraction
------------------------------------------------------------------------*/
static group_votes*
  find_group_record(name,group_list)
char* name;
group_votes *group_list;
{
  group_votes *scan;
  
  scan = group_list;
  while (scan != NULL)
    {
      if (strcasecmp(scan->group_name,name) == 0)
	return scan;
      scan = scan->next_group;
    }
  return NULL;
}
static group_votes *
  create_new_group_record(group_name)
char * group_name;
/* allocate and init a new group record to hold votes for a group */
{
  group_votes *temp;
  temp = allocate(group_votes);
  temp->group_name = (char *) collab_getmemory(strlen(group_name) + 1);
  (void) strcpy(temp->group_name,group_name);
  temp->votes = temp->current_vote = NULL;
  temp->next_group = NULL;
  return temp;
}

static void 
  destroy_groups(g)
group_votes *g;
{
  group_votes *temp;

  while (g != NULL)
    {
      free(g->group_name);
      collab_destroy_votes(g->votes);
      temp = g;
      g = g->next_group;
      free(temp);
    }
}

/*========================================================================
  Random stuff
------------------------------------------------------------------------*/
#ifdef PROVIDE_ISSUE_ERROR
#include <varargs.h>

int
  issue_error(va_alist)
va_dcl
{
  va_list ap;
  char *fmt;

  va_start(ap);
  fmt = va_arg(ap, char *);
  vfprintf(stderr,fmt,ap);
  fprintf(stderr,"\n");
  fflush(stderr);
  va_end(ap);
  return 1;
}
#endif /* PROVIDE_ISSUE_ERROR */

/*========================================================================
  Memory procurement routines
------------------------------------------------------------------------*/
void* 
  collab_getmemory(size)
int size;
{
  void *temp;
  
  temp = (void *) malloc(size);

  if (temp == NULL)
    { /* do something clever... for now, we scream and shout */
      issue_error("Couldn't malloc in colinterface.c");
    }

  return temp;
}

