/* collabtcp.c
     Communicate with a vote server over a tcp link
     by Dave Maltz 10/93

  history:
  10/27/93 created

*/

#include "colinterface.h"
#include "collabdata.h"
#include "collabtcp.h"
#include "colinternal.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef VOTE_SERVER
#include <netdb.h>
#endif /* VOTE_SERVER */
#include <sys/time.h> 		/* needed for select */

#include <stdio.h>
#include <time.h>
#include <memory.h>   /* note: this is going to cause porting problems */
#define BYTE_COPY(to,from,len) (bcopy((from),(to),(len)))
#include <errno.h>


/******* Internal Constants ***********/
#define DEFAULT 0
#define CONNECTED 0x01
#define CREATED 0x02

/******* Internal Globals ***********/
/* name used to connect to the server */
#ifdef VOTE_SERVER
static char *collab_vote_server_name = VOTE_SERVER;
#else
static char *collab_vote_server_name = NULL;
#endif 

static int collab_vote_server_port = VOTE_SERVER_PORT;
/* what port to talk to on the server */

static FILE* serverin = NULL;
static FILE* serverout = NULL;
/* streams associated with the socket which will get connected to the server */

static int serverstatus = DEFAULT;
/* are we attached to the vote server ? */

/****** Internal Procedures *********/
static int read_line_from_server(/* buf, bufsize, length, exception */);
static void dump_serverin_till_cr();

/*========================================================================
  
------------------------------------------------------------------------*/
static void
  drop_connection()
{
  if (serverout != NULL) fclose(serverout);
  if (serverin != NULL) fclose(serverin);
  serverout = serverin = NULL;
  serverstatus = serverstatus & ~CONNECTED;
}

static void
  check_socket_connected()
{
  int sock_rd, sock_wr;  /* fd's for the server socket */
  struct sockaddr_in name;
  int result;
  int c = 0;
#ifdef VOTE_SERVER
  struct hostent *hp, *gethostbyname();
#endif

  if (serverstatus & CONNECTED) return;

  /* int socket(domain, type, protocol) (protocol 6=tcp 0=do the right thing) */
  sock_rd = socket(PF_INET, SOCK_STREAM, 0);

  /** name the socket **/
#ifdef VOTE_SERVER_DQ
  name.sin_addr.s_addr = htonl(VOTE_SERVER_DQ);
#else
  hp = gethostbyname(collab_vote_server_name);
  if (hp == 0) 
    {
      issue_error("couldn't find addr for host %s",collab_vote_server_name);
      return;
    }
  (void) BYTE_COPY((char *) &name.sin_addr, (char *) hp->h_addr_list[0], hp->h_length);
#endif
  name.sin_family = PF_INET;
  name.sin_port = htons(collab_vote_server_port);

  while (c < NUM_RETRIES)
    {
      result = connect(sock_rd, (struct sockaddr *) &name, sizeof(name));
      if (result == 0)
	{ /* we've got a connection, now connect up the file streams */
	  if ((serverin = fdopen(sock_rd, "r")) == NULL) 
	    {
	      close(sock_rd);
	      return;
	    }
	  sock_wr = dup(sock_rd);
	  if ((serverout = fdopen(sock_wr, "w")) == NULL) 
	    {
	      close(sock_wr);
	      fclose(serverin);
	      serverin = NULL;
	      return;
	    }
	  serverstatus = serverstatus | CONNECTED;
	  return;
	}
      c++;
      continue;
    } /* end while not too many retries */
  
  issue_error("couldn't connect to vote server on %s. last errno was %d",
	      collab_vote_server_name, errno);
}

int
  collabtcp_send_buf(buf,len)
char *buf;
int len;
/* send the first len bytes of the buf to the vote server
   returns 1 on success, 0 on failure
*/
{
  int result;
  int c = 0;

  check_socket_connected();
  if ((serverstatus & CONNECTED) == 0)
    {
      return 0;  /* game over folks */
    }

  while (c < NUM_RETRIES)
    {
      /* try to send off the buffer (int fwrite (ptr, size, nitems, stream)) */
      result = fwrite(buf,1,len,serverout);
      fflush(serverout);

      if (result == 0)
	  issue_error("couldn't send data to socket. errno: %d data:%s",
		      errno,buf);
      else if (result < len)
	{
	  buf = buf + result;
	  len = len - result;
	}
      else return 1;

      c++;
    }

  issue_error("repeated attempts to send data failed");
  return 0;
}

#define BUFSIZE 1024

vote_t *  
  collabtcp_get_votes_via_tcp(group_name)
char *group_name;
  /* talk the vote server over the severfd and try to get the votes
     for group_name, returns NULL on failure */
{
  char buf[BUFSIZE];
  vote_t *new_vote,*votes = NULL; /* head of the list of new votes */
  int exception;		/* hold out-of-band data from parse routine */
  int done, len;
  int status; 			/* holds return value of system calls */

  check_socket_connected();	/* make sure we have someone to talk to */
  if ((serverstatus & CONNECTED) == 0)
    {
      return votes;  /* game over folks */
    }

  /* ask the server to send us the votes for the group */
  sprintf(buf,"GROUP %s\n",group_name);
  status = collabtcp_send_buf(buf,strlen(buf));
  if (status == 0) return votes;  /* something bad happened */

  done = 0;
  while (!done)
    {
      status = read_line_from_server(buf,BUFSIZE,&len,&exception);
      if (status != 1)
	{  /* handle exceptions */
	  if (exception == LOST_CONNECTION) return votes;
	  if (exception != NOT_ENUF_BUF) return votes; /* should never happen */
	  /* if not enough buf, set up to parse what we have, and dump the
	     input stream till the next \n */
	  buf[BUFSIZE-1] = '\n';   /* terminate the buffer */
	  dump_serverin_till_cr(); /* dump rest of line */
	}
      /* parse the server's line into a vote */
      new_vote = collab_parse_vote_from_string(buf,&exception);
      if (new_vote != NULL)
	{ /* add the vote onto the list of votes to be returned */
	  new_vote->next_vote = votes;
	  votes = new_vote;
	}	      
      else if (exception == END_OF_DATA)
	{ /* then we're done reading data */
	  done = 1;
	  break;
	}
    } /* end of while !done */
      
  return votes;
}


char *
  collabtcp_get_group_list(num_groups)
int * num_groups;
/* talk to the connected server to get a list of all the groups that have votes
   in them, return this a string with each group sepparated by \n */
{
  int exception;
  char *group_string = NULL;
  char *bufptr, *temp;
  int string_space;	   /* bytes of space allocated for group_string */
  int string_length ;	   /* bytes used by the group_string */
  int status, done, len;
  int templen,oldlen;      /* temps for pointer hacking */
  char buf[10];		   /* there will never be more than 10^10 groups, plez */
  int est_num_groups = 0;  /* the estimated number of groups */
  
  *num_groups = 0; /* in case we bail out on a error condition */

  /* ask server to send us the list of groups */
  (void) strcpy(buf,"AVAIL\n");
  status = collabtcp_send_buf(buf, strlen(buf));
  if (status != 1) return NULL;

  status = read_line_from_server(buf, sizeof(buf), &len, &exception);
  if (status != 1) return NULL;
  est_num_groups = atoi(buf);
  if (est_num_groups == 0) return NULL;

  string_space = est_num_groups * AVE_GRP_NAME_SZ;
  group_string = (char *) collab_getmemory(string_space);
  
  done = 0;
  string_length = 0;
  bufptr = group_string;
  while (!done)
    {
      status = read_line_from_server(bufptr, (string_space - string_length), 
				     &len, &exception);
      if (status == 0)
	{ /* handle exceptions */
	  if (exception != NOT_ENUF_BUF) goto error_return;
	  /* get us more buffer space */
	  string_space = string_space * 2; /* this is way too much memory */
	  temp = (char *) realloc(group_string, string_space); 
	  if (temp == NULL) goto error_return;
	  /* move pointers to new space */
	  group_string = temp;                 
	  bufptr = group_string + string_length;    
	  oldlen = len;
	  templen = string_length + len;
	  /* copy over the rest of the line that didn't make it */
	  status = read_line_from_server((group_string + templen),
					 (string_space - templen),
					 &len, &exception);
	  if (status == 0) goto error_return; /* give up */
	  /* fix up the pointers */
	  len = len + oldlen;
	} /* end of exception handling */

      if (strcmp(bufptr,".\n") == 0)
	{ /* then we're done */
	  *bufptr = '\0';  /* terminate the group string */
	  done = 1;
	  continue;
	}

      /* if we're not done then get ready for the next group name */
      *num_groups = * num_groups + 1;
      bufptr = bufptr + len;
      string_length = string_length + len;
    } /* end while !done */

  return group_string;
  
 error_return:
  free(group_string);
  *num_groups = 0;
  return NULL;
}


/*========================================================================
  Helper routines
------------------------------------------------------------------------*/
static void
  dump_serverin_till_cr()
{
  char c;
  for (; (c = fgetc(serverin)) != '\n';);
}
	
static int
  read_line_from_server(outbuf, bufsize, len, exception)
char *outbuf;
int bufsize;
int *len;
int *exception;
/* fill the outbuf with data from the server until we reach a \n or 
   run out of buffer space, set *len to the length read in */
{	
  
  *exception = NORMAL;
  if (fgets(outbuf, bufsize, serverin) == NULL) 
    {
      drop_connection();
      *exception = LOST_CONNECTION;
      return 0;
    }
  *len = strlen(outbuf);
  if (*len == (bufsize - 1) && outbuf[bufsize-2] != '\n')
    {
      *exception = NOT_ENUF_BUF;
      return 0;
    }
  return 1;
}



