/* filter.c 
   pre-select articles based on some filter criteria 
   for now, use the collaborative voting information */

/* Copyright (c) 1993 Xerox Corporation.  All rights reserved.
 
   Use and copying of this software and preparation of derivative works
   based upon this software are permitted.  Any distribution of this
   software or derivative works must comply with all applicable United
   States export control laws.  This software is made available AS IS, and
   Xerox Corporation makes no warranty about the software, its performance
   or its conformity to any specification. */

#include "config.h"
#include "colinterface.h"
#include "message_id.h"         /* config.h must be included before message_id.h */
#include "filter.h"

#include <string.h>

/******* program global variables ********/
int show_only_filtered_arts = 0; /* global, user visible, boolean flag */
int collab_select_size = 5;        /* >0 => how many arts to select, <=0 => 0 */

/*========================================================================
  this procedure determines how the articles are sorted into an
  ordered list. Tweak this till you like how your filtering works
------------------------------------------------------------------------*/
static int
  vote_compare(v1,v2)
vote_t *v1,*v2;
{
  /* for now, just add up the number of positive votes and subtract the
     number of negative votes
     */
  int score1, score2, temp;
  score1 = v1->vote_data->num_high + 
           v1->vote_data->num_med + 
           v1->vote_data->num_low -
	   v1->vote_data->num_terrible;
  score2 = v2->vote_data->num_high + 
           v2->vote_data->num_med + 
           v2->vote_data->num_low -
	   v2->vote_data->num_terrible;
  temp = score1 - score2;
  if (temp > 0)
    return 1;
  else if (temp < 0)
    return -1;
  else return 0;
}

/*========================================================================
  Get the article information from the collab system and turn it
  into an ordered list
------------------------------------------------------------------------*/
static vote_t *
  sort( unsorted_votes, cmp)
vote_t *unsorted_votes;
int (*cmp)(); /* comparision function, return values like strcmp */
/* do insertion sort on the the votes of the group using comparison funct 
   cmp.  return greatest at the head of list */
{
  vote_t *sorted_votes = NULL;	/* points to head of sorted vote list */
  vote_t *scan_votes;		/* scan over the votes in the unsorted group */
  vote_t *scan_sorted_votes; 	/* scan sorted votes to find insertion point */
  vote_t *prev_vote;		/* vote prior to insertion point */
  vote_t *next_scan_vote;

  scan_votes = unsorted_votes;
  while (scan_votes != NULL)
    {
      scan_sorted_votes = sorted_votes;
      prev_vote = NULL;
      while (scan_sorted_votes != NULL && 
	     (cmp(scan_sorted_votes,scan_votes) == 1))  /* s_s_v > s_v */
	{ /* find the insertion point and the previous vote */
	  prev_vote = scan_sorted_votes;
	  scan_sorted_votes = scan_sorted_votes->next_vote;
	}
      next_scan_vote = scan_votes->next_vote;	/* save the next pointer */
      if (prev_vote == NULL) 
	{ /* elt at head of list or empty list */
	  sorted_votes = scan_votes;
	  sorted_votes->next_vote = scan_sorted_votes; 
	}
      else 
	{ /* standard case */
	  prev_vote->next_vote = scan_votes;
	  scan_votes->next_vote = scan_sorted_votes;
	}
      scan_votes = next_scan_vote;
    }
  return sorted_votes;
}		

static vote_t *
  get_votes_for_group(name)
char *name;
/* returns a list of the votes in order most popular to least popular */
{
  vote_t *votes;

  /* get an unordered list of the votes */
  votes = collab_get_votes_for_group(name);
  
  /* now sort the votes into a list */
  votes = sort(votes, vote_compare);
  return(votes);
}

/*========================================================================
  Perform filtering using collaborative rating information
------------------------------------------------------------------------*/
int
  find_article_with_this_id(articles, n_articles, message_id, signal)
article_header **articles;
int n_articles;
mess_id_type message_id;
int *signal;
{
  int count;
  *signal = NORMAL_EXIT;
  
  for (count = 0 ; count < n_articles ; count++)
    {
      if (mess_id_cmp(message_id, articles[count]->message_id) == 0)
	return count;
    }
  *signal = NOT_FOUND;
}

int
  mark_the_top(num, articles, n_articles, votes, attr_value)
int num;
int n_articles;
article_header **articles;
vote_t *votes;
attr_type attr_value;
{
  /* NOTE: this implementation is not good, the right assumption is probably that
     n_articles is much less than the number of votes in the linked list. a real
     implementation probably puts one of the things in to a hash table */
  
  int count = 0;  /* number of articles selected so far */
  int index;
  vote_t *scan_votes;
  int signal;
  int temp;

  scan_votes = votes;
  while (scan_votes != NULL)
    {
      if (count >= num) break; /* we've found enough articles */

      /* this is stupid hack to keep the collab system from 
	 selecting terrible articles. we cut off the scan down the
	 list of votes once we hit a bad one */
      temp = scan_votes->vote_data->num_high + 
	     scan_votes->vote_data->num_med + 
	     scan_votes->vote_data->num_low -
	     scan_votes->vote_data->num_terrible;
      if (temp < 0) break; /* the rest of the articles were hated... */

      /* note: we need to convert the message id from string form in vote_t
	 to whatever form the message_id has in the articles array */
      index = find_article_with_this_id(articles, n_articles, 
					hash_message_id(scan_votes->message_id),
					&signal);
      if (signal & NOT_FOUND)
	{
	  scan_votes = scan_votes->next_vote;
	  continue;
	}
      articles[index]->attr = attr_value;  /* had been |, now assignment 10/6 */
      count = count + 1;
      scan_votes = scan_votes->next_vote;
    }
  return (count);
}    

int
  filter_do_collaborative_preselect(grp,desired_size,articles,n_articles)
group_header *grp;
int desired_size;
article_header **articles;
int n_articles;
/* 	get a ranked listing of articles 
	and mark the unread articles in the group as selected until
	the group reaches it's desired filtered size */
{
  vote_t *votes;
  int num_marked;
  /* implement a simple caching system for the votes */
  static vote_t *cached_votes = NULL;
  static group_header *cached_grp = NULL;

  /* note: getting the votes might be expensive since we have to sort them.  we
cache them in this procedure. we know when to invalidate the cache b/c grp will
change. when nn exits, the cache will still contain votes for which
collab_destroy_votes won't have been called... */

  if (cached_grp != grp) 
    { /* get the votes for the new group */ 
      if (cached_votes != NULL) collab_destroy_votes(cached_votes); 
      votes = get_votes_for_group(grp->group_name); 
      cached_votes = votes; cached_grp = grp; 
    }
  else votes = cached_votes;

  if (votes == NULL) return -1; /* there's no information to be had */

  num_marked = mark_the_top(desired_size, articles, n_articles, 
			    votes, A_FILTER_SELECT);

  return(num_marked);
}
